#
# (c) 2022, Giorgio Gonnella, University of Goettingen, Germany
#

from schema import Or, Optional, Schema
from loguru import logger
import sys

ARGS_DOC = """\
  --verbose, -v    be verbose
  --version, -V    show script version
  --help, -h       show this help message"""

ARGS_SCHEMA = {Optional("--verbose", default=None): Or(None, True, False),
                                                    Optional(str): object}

def validate(args, *dicts):
  s = ARGS_SCHEMA.copy()
  for d in dicts:
    s.update(d)
  return Schema(s).validate(args)

def setup_logger(verbose):
  logger.remove()
  logger.add(sys.stderr, colorize=sys.stdout.isatty(),
             format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> | "+\
                    "<level>{level}</level> | {message}",
             level="INFO" if verbose else "WARNING")
  return logger



#
# (c) 2020-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Loads the data into the database using the mysql module
"""

import os
import MySQLdb
from pathlib import Path
from .dbloader_sqlwriter import _load_data_sql
from .dbschema import FILE2CLASS

def _connect_and_execute(host, user, passwd, db, unix_socket, statements):
  db = MySQLdb.connect(host=host, user=user, passwd=passwd, db=db,
                       unix_socket=unix_socket, use_unicode=True)
  cursor = db.cursor()
  for statement in statements:
    cursor.execute(statement)
  cursor.close()
  db.commit()

def _load_data(datafile, dbmodel, host, user, passwd, db, unix_socket):
  tablename = dbmodel.__tablename__
  columns = dbmodel.file_column_names()
  _connect_and_execute(host, user, passwd, db, unix_socket,
                       _load_data_sql(datafile, tablename, columns))

def load_all(dumpdir, host, user, passwd, db, unix_socket):
  loaded = []
  for filepfx, klass in FILE2CLASS.items():
    filepath = Path(dumpdir) / f"{filepfx}.dmp"
    if filepath.exists():
      loaded.append((filepfx, filepath))
      _load_data(filepath, klass, host, user, passwd, db, unix_socket)
  return loaded



#
# (c) 2020-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Defines the local database schema
"""

from sqlalchemy.sql import func
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Sequence, Column, Integer, String, \
                       Index, DateTime, Text, Boolean
from sqlalchemy.orm import relationship
from sqlalchemy_repr import PrettyRepresentableBase

# Nt = NCBI Taxonomy

Base = declarative_base(cls=PrettyRepresentableBase)

utf8_cs_args = {'mysql_charset': 'utf8', 'mysql_collate': 'utf8_bin'}

class NtName(Base):
  __tablename__ = 'nt_names'
  time_created = Column(DateTime, server_default=func.now())
  tax_id = Column(Integer, primary_key=True, autoincrement=False, index=True)
  name_txt = Column(String(256), primary_key=True)
  unique_name = Column(String(256), primary_key=True)
  name_class = Column(String(64), primary_key=True)
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["tax_id", "name_txt", "unique_name", "name_class"]

class NtGencode(Base):
  __tablename__ = 'nt_gencode'
  genetic_code_id = Column(Integer, primary_key=True, autoincrement=False)
  time_created = Column(DateTime, server_default=func.now())
  abbreviation = Column(String(64))
  name = Column(String(256))
  cde = Column(String(64))
  starts = Column(String(64))
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["genetic_code_id", "abbreviation", "name", "cde", "starts"]

class NtMerged(Base):
  __tablename__ = 'nt_merged'
  time_created = Column(DateTime, server_default=func.now())
  old_tax_id = Column(Integer, primary_key=True, autoincrement=False)
  new_tax_id = Column(Integer, index=True)
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["old_tax_id", "new_tax_id"]

class NtDivision(Base):
  __tablename__ = 'nt_division'
  division_id = Column(Integer, primary_key=True, autoincrement=False)
  time_created = Column(DateTime, server_default=func.now())
  division_cde = Column(String(3))
  division_name = Column(String(64))
  comments = Column(String(256))
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["division_id", "division_cde", "division_name", "comments"]

class NtDelnode(Base):
  __tablename__ = 'nt_delnodes'
  tax_id = Column(Integer, primary_key=True, autoincrement=False)
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["tax_id"]

class NtCitation(Base):
  __tablename__ = 'nt_citations'
  cit_id = Column(Integer, primary_key=True, autoincrement=False)
  cit_key = Column(String(256))
  pubmed_id = Column(Integer)
  medline_id = Column(Integer)
  url = Column(String(1024))
  text = Column(Text)
  taxid_list = Column(Text(1000000))
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["cit_id", "cit_key", "pubmed_id", "medline_id", "url", "text",
            "taxid_list"]

class NtNode(Base):
  __tablename__ = 'nt_nodes'
  tax_id = Column(Integer, primary_key=True, autoincrement=False)
  parent_tax_id = Column(Integer, index=True)
  rank = Column(String(64), index=True)
  embl_code = Column(String(64))
  division_id = Column(Integer)
  inherited_div_flag = Column(Boolean)
  genetic_code_id = Column(Integer)
  inherited_GC_flag = Column(Boolean)
  mitochondrial_genetic_code_id = Column(Integer)
  inherited_MGC_flag = Column(Boolean)
  GenBank_hidden_flag = Column(Boolean)
  hidden_subtree_root_flat = Column(Boolean)
  comments = Column(String(1024))
  __table_args__ = utf8_cs_args

  @classmethod
  def file_column_names(cls):
    return ["tax_id", "parent_tax_id", "rank", "embl_code", "division_id",
            "inherited_div_flag", "genetic_code_id", "inherited_GC_flag",
            "mitochondrial_genetic_code_id", "inherited_MGC_flag",
            "GenBank_hidden_flag", "hidden_subtree_root_flat", "comments"]

tablename2class = {
  'nt_names': NtName,
  'nt_gencode': NtGencode,
  'nt_merged': NtMerged,
  'nt_division': NtDivision,
  'nt_delnodes': NtDelnode,
  'nt_citations': NtCitation,
  'nt_nodes': NtNode,
}

FILE2CLASS = {
  'names': NtName,
  'gencode': NtGencode,
  'merged': NtMerged,
  'division': NtDivision,
  'delnodes': NtDelnode,
  'citations': NtCitation,
  'nodes': NtNode,
}

DB_MODEL_CLASSES = tablename2class.values()

def create(connection):
  for klass in DB_MODEL_CLASSES:
    klass.metadata.create_all(connection)

def drop(connection):
  for klass in DB_MODEL_CLASSES:
    klass.metadata.drop_all(connection)



#
# (c) 2020-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
Loads the data into the database using SqlAlchemy
"""

from pathlib import Path
from .dbloader_sqlwriter import _load_data_sql
from .dbschema import FILE2CLASS
from sqlalchemy import text

def _load_data(datafile, dbmodel, connection):
  tablename = dbmodel.__tablename__
  columns = dbmodel.file_column_names()
  for line in _load_data_sql(datafile, tablename, columns):
    connection.execute(text(line))

def load_all(dumpdir, connection):
  loaded = []
  for filepfx, klass in FILE2CLASS.items():
    filepath = Path(dumpdir) / f"{filepfx}.dmp"
    if filepath.exists():
      loaded.append((filepfx, filepath))
      _load_data(filepath, klass, connection)
  return loaded



#
# (c) 2020-2022 Giorgio Gonnella, University of Goettingen, Germany
#
"""
SQL statements for loading the data into the database
"""

def _load_data_sql(datafile, tablename, columns):
  """Create SQL script for bulk loading NCBI Taxonomy dump data.

  Non-unique indices are dropped before loading and recomputed afterwards.
  IGNORE is used to handle eventually repeated primary keys.

  Args:
    datafile: NCBI taxonomy dump file
    tablename: table where to insert the data
    columns: table column names in the order of the data in the file

  Returns:
    list of SQL statements
  """
  result = []
  result.append("SET foreign_key_checks = 0;")
  result.append(f"ALTER TABLE {tablename} DISABLE KEYS;")
  sql = f"LOAD DATA LOCAL INFILE '{datafile}' "
  sql += "IGNORE "
  sql += f"INTO TABLE {tablename} "
  sql += r"FIELDS TERMINATED BY '\t|\t' "
  sql += r"LINES TERMINATED BY '\t|\n' "
  sql +="("+",".join(columns)+") "
  result.append(sql)
  result.append(f"ALTER TABLE {tablename} ENABLE KEYS;")
  result.append("SET foreign_key_checks = 1;")
  return result



#
# (c) 2020-2022 Giorgio Gonnella, University of Goettingen, Germany
#

"""
Downloads the NCBI taxonomy dump, if it is newer than the timestamp flag file
"""

import sh
import os

class Downloader():
  REMOTE = "ftp://ftp.ncbi.nih.gov/pub/taxonomy"
  DUMPFILENAME = "taxdump.tar.gz"
  TESTFILENAME = "taxdump_readme.txt"
  TIMESTAMP = "timestamp"

  def __init__(self, outdir):
    self.outdir = outdir
    os.makedirs(self.outdir, exist_ok=True)
    self.testmode = False

  def set_testmode(self):
    self.DUMPFILENAME = self.TESTFILENAME
    self.testmode = True

  def run(self, decompress=True):
    remotefile = self.REMOTE + "/" + self.DUMPFILENAME
    localfile = self.outdir + "/" + self.DUMPFILENAME
    timestampfile = self.outdir + "/" + self.TIMESTAMP
    args = [remotefile, "-o", localfile, "-w", "%{size_download}", "-R"]
    if os.path.exists(timestampfile):
      args += ["-z", timestampfile]
    ret = sh.curl(*args)
    downloaded = int(ret.rstrip())
    if downloaded > 0:
      if decompress and not self.testmode:
        sh.tar("-xzf", localfile, "-C", self.outdir)
        os.unlink(localfile)
      sh.touch(timestampfile)
      return True
    else:
      return False
